import { Host, h } from "@stencil/core";
import { ChartD3Service } from '../../services/chart.d3.service';
import { SeriesType } from '../../types/types';
import { ElementStrategyFactory, SingleLineSeries } from './chart.service';
import { ChartEvents } from './chart.events';
import { COLOR_WHEEL } from '../../services/color-wheel';
let id = 0;
export class Chart {
    constructor() {
        this.id = `anj-chart-${id++}`;
        this.overlayId = `anj-chart-overlay-${id}`;
        this.triedLoadingSeries = false;
        this.service = new SingleLineSeries();
        /**
         * By default, we would like to be able to interact with the popover and it's content.
         * In some other cases we'd prefer to have mouse events penetrate the popover, and forfeit interacting with popover content
         */
        this.disabledPopoverEvents = false;
    }
    componentWillLoad() {
        this.axisConfigs = {
            x: this.service.getDefaultXAxisConfig(),
            y: this.service.getDefaultYAxisConfig()
        };
    }
    componentWillUpdate() {
        this.debounceInitRender();
    }
    async connectedCallback() {
        this.connected = true;
        this.popover = this.element.querySelector('anj-overlay');
        this.popoverComponent = this.element.querySelector('[slot="popover"]');
        // In the event the series call up before chart is connected
        if (this.triedLoadingSeries) {
            this.debounceInitRender();
        }
    }
    disconnectedCallback() {
        this.connected = false;
        this.clearChart();
        if (this.popover) {
            this.popover.close();
        }
        this.popover = null;
        this.popoverComponent = null;
    }
    async onSeriesLoaded() {
        if (this.connected) {
            return this.debounceInitRender();
        }
        else {
            this.triedLoadingSeries = true;
        }
    }
    async onSeriesUpdated({ detail }) {
        this.updateSeries(detail);
    }
    onXAxisLoaded({ target, detail }) {
        const element = target;
        if (element.dimension === 'x') {
            this.xAxisElement = element;
            this.axisConfigs.x = Object.assign({}, this.axisConfigs.x, detail);
        }
        else {
            this.yAxisElement = element;
            this.axisConfigs.y = Object.assign({}, this.axisConfigs.y, detail);
        }
        if (this.connected) {
            this.debounceInitRender();
        }
    }
    async handleResize() {
        this.clearChart();
        return this.debounceCreation();
    }
    // Add Component Events
    addLineEvents(svg, chartOverlay, x, data) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjLineOverlayHoverHandler(svg, data, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    addColumnEvents(svg, chartOverlay, x, data) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjColumnOverlayHoverHandler(svg, data, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    addSliceEvents(slices, data) {
        slices
            .on('mousemove', (_d, i, sliceSVGs) => this.eventsService.anjSVGHoverHandler(sliceSVGs[i], data, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(slices));
    }
    addPointEvents(svgPoints, data, seriesIndex) {
        svgPoints
            .on('mousemove', (_d, i, points) => this.eventsService.anjSVGHoverHandler(points[i], data, i, this.axisConfigs, this.seriesMeta, seriesIndex))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svgPoints));
    }
    addStackedEvents(svg, chartOverlay, x) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjStackedOverlayHoverHandler(svg, this.popoverData, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    // Creation & Rendering
    appendSVG(svg) {
        this.element.appendChild(svg.node());
    }
    async debounceCreation() {
        window.clearTimeout(this.creationTimer);
        return new Promise(resolve => {
            this.creationTimer = window.setTimeout(async () => {
                resolve(this.create());
            }, Chart.creationDebounceLength);
        });
    }
    async debounceInitRender() {
        window.clearTimeout(this.initRenderTimer);
        return new Promise(resolve => {
            this.initRenderTimer = window.setTimeout(async () => {
                resolve((this.connected) ? this.initRender() : null);
            }, Chart.initRenderDebounceLength);
        });
    }
    async create() {
        this.clearChart();
        let x;
        let y;
        const dims = this.service.getDims(this.element);
        // Establish the primary and secondary series
        const hierarchy = ElementStrategyFactory.establishSeriesHierarchy(this.pie, this.stackedColumns, this.columns, this.scatter, this.lines);
        if (!ElementStrategyFactory.validChart(hierarchy.primaryComponents)) {
            console.log('Anj-Chart: No valid data series provided');
            return;
        }
        this.primarySeries = hierarchy.primarySeries;
        this.secondarySeries = hierarchy.secondarySeries;
        this.primaryComponents = hierarchy.primaryComponents;
        this.secondaryComponents = hierarchy.secondaryComponents;
        // Establish the proper service strategy and manipulate the needed properties
        this.service = ElementStrategyFactory.get(this.primarySeries, this.primaryComponents);
        this.multiSeries = this.service.isMultiSeries();
        this.popover.placement = this.service.getPopoverPlacement();
        this.axisConfigs = this.setFormattingOptions(this.service.xAxisTypeOverride(this.axisConfigs));
        // Configure Events Service for hover handling
        this.eventsService = new ChartEvents(this.primarySeries, this.anjHover, this.service, this.popover);
        // Fetch the various formats of data
        const { data, stackedData, minMax } = this.service.formatPrimaryData(this.axisConfigs);
        const flattenedData = this.service.flatten(data);
        // allows for primary and secondary data to be shown together
        this.popoverData = data;
        this.stackedData = stackedData;
        // Base for Chart
        const svg = ChartD3Service.createSVG(dims);
        // Attach Chart to the DOM
        this.appendSVG(svg);
        // Add axis for charts that require them
        if (this.service.usesAxis()) {
            // Y-Axis
            y = (this.yAxisElement)
                // Create defined x axis from anj-x-axis slot
                ? await this.hydrateYAxis(dims, flattenedData, svg, minMax)
                // Create hidden x axis for continuity
                : ChartD3Service.addYAxis(svg, dims, flattenedData, this.axisConfigs, minMax);
            // X-Axis
            x = (this.xAxisElement)
                // Create defined x axis from anj-x-axis slot
                ? await this.hydrateXAxis(dims, flattenedData, svg, this.primarySeries, y)
                // Create hidden x axis for continuity
                : ChartD3Service.addXAxis(svg, dims, flattenedData, this.axisConfigs, undefined, this.primarySeries, y);
        }
        // Components
        await this.hydratePrimaryComponents(svg, x, y, data, dims);
        await this.hydrateSecondaryComponents(svg, x, y, dims);
        return svg;
    }
    // Hydrate slotted components
    async hydratePrimaryComponents(svg, x, y, data, dims) {
        switch (this.primarySeries) {
            case SeriesType.Column:
                const x1 = ChartD3Service.getX1(x, data, this.multiSeries);
                return this.hydrateColumns(svg, data, x, x1, y, dims);
            case SeriesType.Stacked:
                return this.hydrateStackedColumns(svg, this.stackedData, x, y, dims);
            case SeriesType.Line:
                return this.hydrateLinesPrimary(svg, data, x, y, dims);
            case SeriesType.Scatter:
                return this.hydrateScatter(svg, data, x, y, dims);
            case SeriesType.Pie:
                return this.hydratePie(svg, data, dims);
            default:
                return;
        }
    }
    async hydrateSecondaryComponents(svg, x, y, dims) {
        if (this.secondarySeries) {
            const lineData = this.secondaryComponents.map(l => l.data);
            const formattedLineData = this.service.coreFormatData(this.axisConfigs, lineData);
            return this.hydrateLinesSecondary(svg, formattedLineData, x, y, dims);
        }
        else {
            return Promise.resolve();
        }
    }
    async hydrateColumns(svg, data, x, x1, y, dims) {
        this.seriesMeta = [];
        await Promise.all(this.columns.map((c, i) => {
            this.seriesMeta[i] = { name: c.name, color: undefined };
            return c.init(svg, x, y, data[i], i, dims, x1);
        }));
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addColumnEvents(svg, chartOverlay, x, data);
    }
    async hydrateLinesPrimary(svg, data, x, y, dims) {
        this.seriesMeta = [];
        if (this.lines.length === 1) {
            const renderArea = !data[0].some(point => point.y < 0);
            this.lines[0].area = renderArea;
        }
        const colors = await Promise.all(this.lines.map((l, i) => l.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries())));
        this.seriesMeta = colors.map((color, i) => {
            return { name: this.lines[i].name, color };
        });
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addLineEvents(svg, chartOverlay, x, data);
        // Animated L=>R reveal
        if (!this.service.isMultiSeries()) {
            ChartD3Service.addRevealAnimation(svg, dims, this.axisConfigs, data);
        }
    }
    async hydrateLinesSecondary(svg, data, x, y, dims) {
        const sIndex = this.primaryComponents.length;
        const colors = await Promise.all(this.lines.map((l, i) => l.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries(), sIndex)));
        colors.forEach((color, i) => {
            this.popoverData.push(data[i]);
            const popoverIndex = i + this.seriesMeta.length;
            this.seriesMeta[popoverIndex] = { name: this.lines[i].name, color };
        });
    }
    async hydrateScatter(svg, data, x, y, dims) {
        this.seriesMeta = [];
        const points = await Promise.all(this.scatter.map((b, i) => {
            return b.init(svg, x, y, data[i], i, dims);
        }));
        points.forEach((b, i) => {
            this.seriesMeta[i] = { name: b.name, color: b.color };
            this.addPointEvents(b.points, data[i], i);
        });
        // Animated L=>R reveal
        ChartD3Service.addRevealAnimation(svg, dims, this.axisConfigs, data);
    }
    async hydratePie(svg, data, dims) {
        const { slices, seriesMeta } = await this.pie[0].init(svg, undefined, undefined, data[0], 0, dims);
        this.seriesMeta = seriesMeta;
        this.addSliceEvents(slices, data[0]);
    }
    async hydrateStackedColumns(svg, data, x, y, dims) {
        this.seriesMeta = [];
        const rendered = await Promise.all(this.stackedColumns.map((c, i) => c.init(svg, x, y, data[i], i, dims)));
        this.seriesMeta = rendered.map(({ color }, i) => {
            return { name: this.stackedColumns[i].name, color };
        });
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addStackedEvents(svg, chartOverlay, x);
    }
    hydrateXAxis(dims, data, svg, primarySeries, y) {
        return this.xAxisElement.init(svg, dims, data, this.axisConfigs, undefined, primarySeries, y);
    }
    hydrateYAxis(dims, data, svg, minMax) {
        return this.yAxisElement.init(svg, dims, data, this.axisConfigs, minMax, undefined);
    }
    // Load Slotted Components
    async initRender() {
        this.clearChart();
        const [columns, scatter, stackedColumns, lines, pie] = await Promise.all([
            this.loadSeries('column'),
            this.loadSeries('scatter'),
            this.loadSeries('stacked-column'),
            this.loadSeries('line'),
            this.loadSeries('pie')
        ]);
        this.columns = columns.length > 0 ? columns : undefined;
        this.pie = pie.length > 0 ? pie : undefined;
        this.scatter = scatter.length > 0 ? scatter : undefined;
        this.stackedColumns = stackedColumns.length > 0 ? stackedColumns : undefined;
        this.lines = lines.length > 0 ? lines : undefined;
        return this.create();
    }
    loadSeries(type) {
        return Promise.all(Array.from(this.element.querySelectorAll(`anj-series[type="${type}"]`))
            .map((col) => col.componentOnReady()));
    }
    updateSeries(seriesInfo) {
        if (seriesInfo.type === SeriesType.Pie) {
            const data = this.service.coreFormatData(this.axisConfigs, seriesInfo.newData.data);
            this.addSliceEvents(seriesInfo.slices, data);
            this.seriesMeta = seriesInfo.newData.data.map((d, i) => ({
                color: d.color || COLOR_WHEEL[i],
                name: d.name
            }));
        }
    }
    setFormattingOptions(axisConfigs) {
        const formattingOptions = {
            locale: this.locale,
            currency: this.currency
        };
        axisConfigs.x.formattingOptions = formattingOptions;
        axisConfigs.y.formattingOptions = formattingOptions;
        return axisConfigs;
    }
    clearChart() {
        const svg = this.element.querySelector('svg');
        if (svg) {
            svg.remove();
        }
    }
    overlayStyle() {
        return {
            'pointer-events': (this.disabledPopoverEvents) ? 'none' : 'all'
        };
    }
    render() {
        const ids = {
            chartId: this.id,
            overlayId: this.overlayId
        };
        return (h(Host, { id: this.id },
            h("anj-overlay", { id: this.overlayId, style: this.overlayStyle(), className: "anj-popover", offset: "0, 15", ref: r => this.popover = r, onMouseMove: () => this.eventsService.popoverHoverStart(), onMouseOut: (e) => this.eventsService.popoverHoverEnd(e, ids) },
                h("slot", { name: "popover" }),
                this.popoverComponent ? undefined : this.multiSeries && !this.columns ?
                    h("anj-popover-content-multi", null) : h("anj-popover-content-single", null),
                (this.columns || !this.multiSeries) ? h("anj-arrow", null) : undefined)));
    }
    static get is() { return "anj-chart"; }
    static get originalStyleUrls() { return {
        "$": ["chart.scss"]
    }; }
    static get styleUrls() { return {
        "$": ["chart.css"]
    }; }
    static get properties() { return {
        "currency": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": ""
            },
            "attribute": "currency",
            "reflect": false
        },
        "locale": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": ""
            },
            "attribute": "locale",
            "reflect": false
        },
        "disabledPopoverEvents": {
            "type": "boolean",
            "mutable": false,
            "complexType": {
                "original": "boolean",
                "resolved": "boolean",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "By default, we would like to be able to interact with the popover and it's content.\nIn some other cases we'd prefer to have mouse events penetrate the popover, and forfeit interacting with popover content"
            },
            "attribute": "disabled-popover-events",
            "reflect": false,
            "defaultValue": "false"
        }
    }; }
    static get states() { return {
        "multiSeries": {},
        "popoverComponent": {}
    }; }
    static get events() { return [{
            "method": "anjHover",
            "name": "anjHover",
            "bubbles": true,
            "cancelable": true,
            "composed": true,
            "docs": {
                "tags": [],
                "text": "Emitted when the use hovers over a data point or series of them"
            },
            "complexType": {
                "original": "CoordinateData[]",
                "resolved": "CoordinateData[]",
                "references": {
                    "CoordinateData": {
                        "location": "import",
                        "path": "../../types/types"
                    }
                }
            }
        }]; }
    static get elementRef() { return "element"; }
    static get listeners() { return [{
            "name": "anjSeriesDidLoad",
            "method": "onSeriesLoaded",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "anjSeriesDidUpdate",
            "method": "onSeriesUpdated",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "anjAxisDidLoad",
            "method": "onXAxisLoaded",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "resize",
            "method": "handleResize",
            "target": "window",
            "capture": false,
            "passive": true
        }]; }
}
Chart.creationDebounceLength = 50;
Chart.initRenderDebounceLength = 100;
